#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

"""Extends the standard `types` module.

Standard module defines names for all type symbols known in the standard
interpreter.

This modules mirrors all the functions (and, in general, objects) from the
standard library module `types`:mod:; but it also includes several new types
and type-related functions.

In Jython and PyPy, `MemberDescriptorType` is identical to
`GetSetDescriptorType`; to mantain compatibility in some `xotl.tools` code,
they are differentiated in this module.

"""

import re
from inspect import _static_getmro  # type: ignore
from types import (
    BuiltinFunctionType,
    BuiltinMethodType,
    FunctionType,
    GetSetDescriptorType,
    LambdaType,
    MemberDescriptorType,
    MethodType,
)
from typing import TypeVar

from typing_extensions import Protocol
from xotl.tools.params import Undefined
from xotl.tools.symbols import Unset as _unset

try:
    from types import NoneType  # type: ignore
except ImportError:
    try:
        # In PyPy3 'NoneType' is a built-in
        from builtins import NoneType  # type: ignore
    except ImportError:
        NoneType = type(None)  # type: ignore

EllipsisType = type(Ellipsis)


# Check Jython and PyPy peculiarity
if MemberDescriptorType is GetSetDescriptorType:

    class _foo:
        __slots__ = "bar"

    MemberDescriptorType = type(_foo.bar)  # type: ignore
    del _foo


FuncTypes = tuple({
    FunctionType,
    MethodType,
    LambdaType,
    BuiltinFunctionType,
    BuiltinMethodType,
})

func_types = FuncTypes  # Just an alias


RegexPattern = type(re.compile(""))


def _get_mro_attr(target, name, default=Undefined, /):
    """Get a named attribute from a type.

    Similar to `getattr` but looking in the MRO dictionaries for the type.
    Used internally in this module.

    For example::

      >>> class A(SimpleNamespace):
      ...     x = 12
      ...     y = 34

      >>> class B(A):
      ...     y = 56

      >>> b = B(x=1, y=2)

      >>> _get_mro_attr(b, 'x')
      12

      >>> _get_mro_attr(b, 'y')
      56

    """
    # force type
    target = target if isinstance(target, type) else type(target)
    target_mro = _static_getmro(target)
    cls = next((c for c in target_mro if name in c.__dict__), _unset)
    if cls is not _unset:
        return cls.__dict__[name]
    elif default is not Undefined:
        return default
    else:
        msg = "'{}' type has no attribute '{}'"
        raise AttributeError(msg.format(target, name))


def is_staticmethod(cls, name):
    """Returns true if a `method` is a static method.

    :param cls: The class or object that holds the method.

    :param name: The name of the method.

    When a static-method is declared, you can not test that condition using
    the traditional way::

      >>> class Foo:
      ...     @staticmethod
      ...     def bar():
      ...         pass

      >>> isinstance(Foo.bar, staticmethod)
      False

    Using this function::

      >>> is_staticmethod(Foo, 'bar')
      True

    """
    desc = _get_mro_attr(cls, name)
    return isinstance(desc, staticmethod)


def is_classmethod(cls, name):
    """Returns if a `method` is a class method.

    :param cls: The class or object that holds the method.

    :param name: The name of the method.

    When a class-method is declared, you can not test that condition using the
    traditional way::

      >>> class Foo:
      ...     @classmethod
      ...     def bar(cls):
      ...         pass

      >>> isinstance(Foo.bar, classmethod)
      False

    Using this function::

      >>> is_classmethod(Foo, 'bar')
      True

    """
    desc = _get_mro_attr(cls, name)
    return isinstance(desc, classmethod)


def is_instancemethod(cls, name):
    """Returns if a `method` is neither a static nor a class method.

    :param cls: The class or object that holds the method.

    :param name: The name of the method.

    To find out if a method is "normal", ``isinstance(obj.method,
    MethodType)`` can't be used::

      >>> class Foobar:
      ...     @classmethod
      ...     def cm(cls):
      ...         pass
      ...     def im(self):
      ...         pass

      >>> isinstance(Foobar.cm, MethodType)
      True

    Using this function::

      >>> is_instancemethod(Foobar, 'im')
      True

      >>> is_instancemethod(Foobar, 'cm')
      False

    """
    desc = _get_mro_attr(cls, name)
    return isinstance(desc, FunctionType)  # noqa


class TypeClass(Protocol):
    pass


# fmt: off
class EqTypeClass(TypeClass, Protocol):  # pragma: no cover
    def __eq__(self, other) -> bool: ...  # noqa
    def __ne__(self, other) -> bool: ...  # noqa


class OrdTypeClass(EqTypeClass, Protocol):  # pragma: no cover
    def __le__(self, other) -> bool: ...  # noqa
    def __lt__(self, other) -> bool: ...  # noqa
    def __ge__(self, other) -> bool: ...  # noqa
    def __gt__(self, other) -> bool: ...  # noqa
# fmt: on


TEq = TypeVar("TEq", bound=EqTypeClass)
TOrd = TypeVar("TOrd", bound=OrdTypeClass)


__all__ = (
    "EllipsisType",
    "EqTypeClass",
    "FuncTypes",
    "MemberDescriptorType",
    "NoneType",
    "OrdTypeClass",
    "RegexPattern",
    "TEq",
    "TOrd",
    "TypeClass",
    "func_types",
    "is_classmethod",
    "is_instancemethod",
    "is_staticmethod",
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

import pytest
from hypothesis import given, settings, strategies
from xotl.tools.future.datetime import (
    DateTimeSpan,
    EmptyTimeSpan,
    TimeSpan,
    date,
    daterange,
    datetime,
    timedelta,
)
from xotl.tools.testing.datetime import datetimespans, timespans

dates = strategies.dates
maybe_date = dates() | strategies.none()
maybe_datetimes = strategies.datetimes() | strategies.none()


def test_datetime_imports():
    from xotl.tools.future import datetime
    from xotl.tools.future import datetime as dt

    assert datetime is dt

    from xotl.tools.future.datetime import TimeSpan
    from xotl.tools.future.datetime import TimeSpan as TS

    assert TS is TimeSpan


def test_daterange_stop_only():
    result = list(daterange(date(1978, 10, 21)))
    assert result[0] == date(1978, 10, 1)
    assert result[-1] == date(1978, 10, 20)


def test_daterange_empty():
    assert [] == list(daterange(date(1978, 10, 21), -2))
    assert [] == list(daterange(date(1978, 10, 21), date(1978, 10, 10)))
    assert [] == list(daterange(date(1978, 10, 10), date(1978, 10, 20), -1))


def test_daterange_going_back_in_time():
    result = list(daterange(date(1978, 10, 21), -2, -1))
    assert result[0] == date(1978, 10, 21)
    assert result[-1] == date(1978, 10, 20)


def test_daterange_invalid_int_stop():
    with pytest.raises(TypeError):
        daterange(10)


def test_daterange_invalid_step():
    with pytest.raises(ValueError):
        daterange(None, date(1978, 10, 21), 0)


@given(timespans(), timespans())
def test_intersection_commutable(ts1, ts2):
    # Commutable
    assert ts2 * ts1 == (ts1 & ts2)


@given(timespans(), timespans())
def test_intersection_containment(ts1, ts2):
    overlap = ts1 * ts2
    if overlap is not None:
        # The intersection must always be totally covered by both ts1 and ts2,
        # unless ts1 and ts2 don't intersect
        assert (overlap & ts1) == overlap
        assert (overlap <= ts1) is True
        assert (overlap <= ts2) is True


@given(timespans(), timespans())
def test_comparision(ts1, ts2):
    if ts1 <= ts2 <= ts1:
        assert ts1 == ts2
    if ts1 == ts2:
        assert ts1 <= ts2 <= ts1

    # Single day intersection and equality test
    if ts1.start_date:
        assert ts1 * ts1.start_date == ts1.start_date
        assert ts1.start_date * ts1 == ts1.start_date
        assert ts1.start_date in ts1


@given(timespans(), datetimespans())
@settings(deadline=500)
def test_interaction_timespan_with_datetimespans(ts, dts):
    assert isinstance(dts & ts, (DateTimeSpan, type(EmptyTimeSpan)))
    assert isinstance(ts & dts, (DateTimeSpan, type(EmptyTimeSpan)))

    assert DateTimeSpan.from_timespan(ts) == ts
    assert ts == DateTimeSpan.from_timespan(ts)

    ts = TimeSpan("2018-01-01", "2018-01-01")
    dts = DateTimeSpan("2018-01-01 00:00:01", "2018-01-01 00:00:02")

    assert ts != dts
    assert dts != ts


@given(timespans(), timespans(), dates())
def test_general_cmp_properties(ts1, ts2, date):
    assert bool(ts1 <= ts2) == bool(ts2 >= ts1)
    # In Python 2, dates have a __le__ that does no compare to timespans.
    assert bool(TimeSpan.from_date(date) <= ts2) == bool(ts2 >= date)

    overlap = ts1 & ts2
    if not overlap:
        # Disjoint sets are not orderable...
        assert not (ts1 <= ts2) and not (ts2 <= ts1)


@given(timespans(unbounds="future"))
def test_outside_date(ts):
    from datetime import timedelta

    assert ts.start_date
    outsider = ts.start_date - timedelta(1)
    assert outsider not in ts


@given(datetimespans(dates=strategies.datetimes(min_value=datetime(1, 1, 2)), unbounds="future"))
def test_outside_datetime(dts):
    from datetime import timedelta

    assert dts.start_date
    outsider = dts.start_date - timedelta(1)
    assert outsider not in dts


@given(timespans() | datetimespans())
def test_empty_timespan(ts):
    assert ts >= EmptyTimeSpan <= ts, "Empty is a subset of any TS"

    assert EmptyTimeSpan <= EmptyTimeSpan >= EmptyTimeSpan, "Empty is a subset of itself"

    assert not EmptyTimeSpan, "Empty is considered False"

    assert not (ts <= EmptyTimeSpan), "Empty is not a superset of any TS"

    type(EmptyTimeSpan)() is EmptyTimeSpan

    assert EmptyTimeSpan & ts == EmptyTimeSpan * ts == EmptyTimeSpan
    assert EmptyTimeSpan | ts == EmptyTimeSpan + ts == ts


@given(timespans(unbounds="none"), timespans())
def test_failure_of_triple_intersection(ts1, ts2):
    from datetime import timedelta

    ts0 = TimeSpan.from_date(ts1.start_date - timedelta(1))
    assert not (ts0 & ts1 & ts2)


@given(strategies.dates())
def test_dates_are_representable(value):
    from xotl.tools.future.datetime import date

    class mydate(date):
        pass

    value = mydate(value.year, value.month, value.day)
    assert value.strftime("%Y-%m-%d")


@given(timespans(unbounds="none"))
def test_timespans_are_representable(value):
    assert repr(value)


@given(timespans(unbounds="none"))
def test_generate_valid_timespans(ts):
    assert ts.valid


@given(timespans(unbounds="none"))
def test_ts_returns_dates_not_subtypes(ts):
    from datetime import date

    assert type(ts.start_date) is date
    assert type(ts.end_date) is date


@given(timespans(unbounds="none"), strategies.dates())
def test_operate_with_timespans(ts, d):
    assert ts.start_date - d is not None
    assert d - ts.start_date is not None


@given(timespans(unbounds="none"), timespans(unbounds="none"))
def test_definition_of_overlaps(ts1, ts2):
    assert ts1.overlaps(ts2) == bool(ts1 & ts2)


@given(timespans(unbounds="none"))
def test_duplication_of_timespans(ts1):
    ts2 = TimeSpan(ts1.start_date, ts1.end_date)
    assert {ts1, ts2} == {ts1}, "ts1 and ts2 are equal but different!"


@given(timespans(), strategies.integers(min_value=-1000, max_value=1000))
def test_timespans_displacement_reversed(ts1, delta):
    try:
        assert (ts1 << delta) == (ts1 >> -delta)
    except OverflowError:
        pass


@given(timespans(), strategies.integers(min_value=-1000, max_value=1000))
def test_timespans_displacement_keeps_unbounds(ts1, delta):
    try:
        assert ts1.unbound == (ts1 << delta).unbound
        assert ts1.future_unbound == (ts1 << delta).future_unbound
        assert ts1.past_unbound == (ts1 << delta).past_unbound
    except OverflowError:
        pass


@given(timespans(), strategies.integers(min_value=-1000, max_value=1000))
def test_timespans_displacement_doubled(ts1, delta):
    try:
        assert ((ts1 << delta) << delta) == (ts1 << (2 * delta))
    except OverflowError:
        pass


@given(timespans(), strategies.integers(min_value=-1000, max_value=1000))
def test_timespans_displacement_backandforth(ts1, delta):
    try:
        assert ts1 == ((ts1 << delta) >> delta) == (ts1 << 0) == (ts1 >> 0)
    except OverflowError:
        pass


@given(timespans(unbounds="none"), strategies.integers(min_value=-1000, max_value=1000))
def test_timespans_displacement_dates(ts1, delta):
    try:
        res = ts1 << delta
    except OverflowError:
        # Ignore if the date it's being displaced to non-supported date,
        # that's for the client to deal with
        pass
    else:
        assert (res.start_date - ts1.start_date).days == -delta
        assert (res.end_date - ts1.end_date).days == -delta
    try:
        res = ts1 >> delta
    except OverflowError:
        # Ignore if the date it's being displaced to non-supported date,
        # that's for the client to deal with
        pass
    else:
        assert (res.start_date - ts1.start_date).days == delta
        assert (res.end_date - ts1.end_date).days == delta


@given(
    timespans(unbounds="none") | datetimespans(unbounds="none"),
    strategies.integers(min_value=-1000, max_value=1000),
)
def test_timespans_displacement_keeps_the_len(ts1, delta):
    try:
        res = ts1 << delta
    except OverflowError:
        pass
    else:
        assert len(ts1) == len(res)


@given(timespans() | datetimespans())
def test_timespans_are_pickable(ts):
    import pickle

    for proto in range(1, pickle.HIGHEST_PROTOCOL + 1):
        assert ts == pickle.loads(pickle.dumps(ts, proto))


def test_empty_timespan_is_pickable():
    import pickle

    for proto in range(1, pickle.HIGHEST_PROTOCOL + 1):
        assert EmptyTimeSpan is pickle.loads(pickle.dumps(EmptyTimeSpan, proto))


@given(strategies.datetimes(), strategies.datetimes())
def test_timespan_with_datetimes(d1, d2):
    from datetime import date as d
    from datetime import datetime as dt

    ts = TimeSpan(d1, d2)
    assert not isinstance(ts.start_date, dt)
    assert isinstance(ts.start_date, d)
    assert not isinstance(ts.end_date, dt)
    assert isinstance(ts.end_date, d)


@given(datetimespans())
def test_datetimespans_ts_fields(dts):
    if dts.start_datetime is None:
        assert dts.start_date is None
    else:
        assert dts.start_date == dts.start_datetime.date()
    if dts.end_datetime is None:
        assert dts.end_date is None
    else:
        assert dts.end_date == dts.end_datetime.date()


@given(maybe_date, datetimespans())
def test_setting_start_date(d, dts):
    dts.start_date = d
    if d is not None:
        assert dts.start_datetime.date() == d
        assert dts.start_datetime == datetime(d.year, d.month, d.day)
    else:
        assert dts.start_datetime is None


@given(maybe_date, datetimespans())
def test_setting_end_date(d, dts):
    dts.end_date = d
    if d is not None:
        assert dts.end_datetime.date() == d
        assert dts.end_datetime == datetime(d.year, d.month, d.day, 23, 59, 59)
    else:
        assert dts.end_datetime is None


@given(
    strategies.datetimes(min_value=datetime(1970, 1, 1), max_value=datetime(5000, 12, 31)),
    strategies.integers(min_value=1, max_value=200),
)
def test_timespan_diff(start_date, delta):
    days = timedelta(days=delta)
    big = TimeSpan(start_date, start_date + 3 * days)
    x, y = big.diff(big)
    assert x is EmptyTimeSpan and y is EmptyTimeSpan

    x, y = big.diff(EmptyTimeSpan)
    assert x == big and y is EmptyTimeSpan

    outsider = big >> 4 * days
    assert not big & outsider
    x, y = big.diff(outsider)
    assert x == big and y is EmptyTimeSpan
    x, y = outsider.diff(big)
    assert x == outsider and y is EmptyTimeSpan

    atstart = TimeSpan(start_date, start_date + days)
    assert atstart < big
    x, y = big.diff(atstart)
    assert x is EmptyTimeSpan
    assert y
    assert add_timespans(atstart, y) == big

    beforestart = TimeSpan(start_date - days, start_date + days)
    x, y = big.diff(beforestart)
    assert x is EmptyTimeSpan
    assert y
    assert add_timespans(beforestart & big, y) == big

    atend = TimeSpan(start_date + 2 * days, start_date + 3 * days)
    assert atend < big
    x, y = big.diff(atend)
    assert y is EmptyTimeSpan
    assert x
    assert add_timespans(x, atend) == big

    afterend = TimeSpan(start_date + 2 * days, start_date + 4 * days)
    x, y = big.diff(afterend)
    assert y is EmptyTimeSpan
    assert x
    assert add_timespans(x, big & afterend) == big

    middle = TimeSpan(start_date + days, start_date + 2 * days)
    assert middle < big
    x, y = big.diff(middle)
    assert x and y
    assert add_timespans(add_timespans(x, middle), y) == big

    bigger = TimeSpan(start_date - days, start_date + 4 * days)
    x, y = big.diff(bigger)
    assert x is EmptyTimeSpan and y is EmptyTimeSpan


@given(
    strategies.datetimes(min_value=datetime(1970, 1, 1), max_value=datetime(5000, 12, 31)),
    strategies.integers(min_value=1, max_value=10000),
)
@settings(deadline=None)
def test_datetimespan_diff(start_date, delta):
    secs = timedelta(seconds=delta)
    big = DateTimeSpan(start_date, start_date + 3 * secs)
    assert big.start_datetime == start_date
    assert big.end_datetime == start_date + 3 * secs
    x, y = big.diff(big)
    assert x is EmptyTimeSpan and y is EmptyTimeSpan

    x, y = big.diff(EmptyTimeSpan)
    assert x == big and y is EmptyTimeSpan

    outsider = big >> 4 * secs
    assert not big & outsider
    x, y = big.diff(outsider)
    assert x == big and y is EmptyTimeSpan
    x, y = outsider.diff(big)
    assert x == outsider and y is EmptyTimeSpan

    atstart = DateTimeSpan(start_date, start_date + secs)
    assert atstart < big
    x, y = big.diff(atstart)
    assert x is EmptyTimeSpan
    assert y
    assert add_dtspans(atstart, y) == big

    beforestart = DateTimeSpan(start_date - secs, start_date + secs)
    x, y = big.diff(beforestart)
    assert x is EmptyTimeSpan
    assert y
    assert add_dtspans(beforestart & big, y) == big

    atend = DateTimeSpan(start_date + 2 * secs, start_date + 3 * secs)
    assert atend < big
    x, y = big.diff(atend)
    assert y is EmptyTimeSpan
    assert x
    assert add_dtspans(x, atend) == big

    afterend = DateTimeSpan(start_date + 2 * secs, start_date + 4 * secs)
    x, y = big.diff(afterend)
    assert y is EmptyTimeSpan
    assert x
    assert add_dtspans(x, big & afterend) == big

    middle = DateTimeSpan(start_date + secs, start_date + 2 * secs)
    assert middle < big
    x, y = big.diff(middle)
    assert x and y
    assert add_dtspans(add_dtspans(x, middle), y) == big

    bigger = DateTimeSpan(start_date - secs, start_date + 4 * secs)
    x, y = big.diff(bigger)
    assert x is EmptyTimeSpan and y is EmptyTimeSpan


@given(datetimespans())
def test_datetimespan_repr(ts):
    assert eval(repr(ts)) == ts
    assert DateTimeSpan("2018-01-01") == DateTimeSpan(datetime(2018, 1, 1))


def add_timespans(x, y):
    if x.end_date == y.start_date - timedelta(1):
        return TimeSpan(x.start_date, y.end_date)
    else:
        raise ValueError


def add_dtspans(x, y):
    if x.end_datetime == y.start_datetime - timedelta(seconds=1):
        return DateTimeSpan(x.start_datetime, y.end_datetime)
    else:
        raise ValueError



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
import unittest


def test_iscollection():
    # TODO: move this test to equivalent for `xotl.tools.values.simple.logic_collection_coerce`
    from collections import UserDict, UserList

    def is_collection(arg):
        from collections.abc import Iterable, Mapping

        avoid = (Mapping, str)
        return isinstance(arg, Iterable) and not isinstance(arg, avoid)

    assert is_collection("all strings are iterable") is False
    assert is_collection(1) is False
    assert is_collection(range(1)) is True
    assert is_collection({}) is False
    assert is_collection(tuple()) is True
    assert is_collection(set()) is True
    assert is_collection(a for a in range(100)) is True

    class Foobar(UserList):
        pass

    assert is_collection(Foobar()) is True

    class Foobar(UserDict):
        pass

    assert is_collection(Foobar()) is False


class NoneTypeTests(unittest.TestCase):
    "To avoid FlyCheck errors"

    def test_identity(self):
        from xotl.tools.future.types import NoneType

        self.assertIs(NoneType, type(None))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

import copy
import pickle
import sys
import unittest
from collections.abc import MutableMapping
from random import shuffle

from xotl.tools.future.collections import DefaultDict, RankedDict, defaultdict


class TestDefaultDict(unittest.TestCase):
    def test_defaultdict(self):
        d = DefaultDict(lambda key, _: "a")
        self.assertEqual("a", d["abc"])
        d["abc"] = 1
        self.assertEqual(1, d["abc"])

    def test_defaultdict_clone(self):
        d = DefaultDict(lambda key, d: d["a"], {"a": "default"})
        self.assertEqual("default", d["abc"])

        d = DefaultDict(lambda key, d: d[key])
        with self.assertRaises(KeyError):
            d["abc"]


class TestDefaultDictAlias(unittest.TestCase):
    def test_defaultdict(self):
        d = defaultdict(lambda key, _: "a")
        self.assertEqual("a", d["abc"])
        d["abc"] = 1
        self.assertEqual(1, d["abc"])

    def test_defaultdict_clone(self):
        d = defaultdict(lambda key, d: d["a"], {"a": "default"})
        self.assertEqual("default", d["abc"])

        d = defaultdict(lambda key, d: d[key])
        with self.assertRaises(KeyError):
            d["abc"]


def test_stacked_dict_with_newpop():
    """Test that stacked.pop has the same semantics has dict.pop."""
    from xotl.tools.future.collections import StackedDict

    sd = StackedDict(a="level-0", b=1)
    assert sd.pop("a") == "level-0"
    assert sd.pop("non", sd) is sd
    try:
        sd.pop("non")
    except KeyError:
        pass
    else:
        assert False, "Should have raised a KeyError"


def test_stacked_dict():
    from xotl.tools.future.collections import StackedDict

    sd = StackedDict(a="level-0")
    assert sd.peek() == dict(a="level-0")
    sd.push_level(a=1, b=2, c=10)
    assert sd.level == 1
    assert sd.peek() == dict(a=1, b=2, c=10)
    sd.push_level(b=4, c=5)
    assert sd.peek() == dict(b=4, c=5)
    assert sd.level == 2
    assert sd["b"] == 4
    assert sd["a"] == 1
    assert sd["c"] == 5
    assert len(sd) == 3
    del sd["c"]
    try:
        del sd["c"]
        assert False, "Should have raise KeyError"
    except KeyError:
        pass
    except Exception:
        assert False, "Should have raise KeyError"
    assert sd.pop_level() == {"b": 4}
    assert sd["b"] == 2
    assert sd["a"] == 1
    assert len(sd) == 3
    sd.pop_level()
    assert sd["a"] == "level-0"
    try:
        sd.pop_level()
        assert False, "Level 0 cannot be poped. " "It should have raised a TypeError"
    except TypeError:
        pass
    except Exception:
        assert False, "Level 0 cannot be poped. " "It should have raised a TypeError"


class TestRankedDict(unittest.TestCase):
    def test_init(self):
        with self.assertRaises(TypeError):
            # too many args
            RankedDict([("a", 1), ("b", 2)], None)
        pairs = [("a", 1), ("b", 2), ("c", 3), ("d", 4), ("e", 5)]
        # dict input
        self.assertEqual(sorted(RankedDict(dict(pairs)).items()), pairs)
        # kwds input
        self.assertEqual(sorted(RankedDict(**dict(pairs)).items()), pairs)
        # pairs input
        self.assertEqual(list(RankedDict(pairs).items()), pairs)
        # mixed input
        self.assertNotEqual(
            list(RankedDict([("a", 1), ("b", 2), ("c", 9), ("d", 4)], c=3, e=5).items()),
            pairs,
        )

        # Make sure that direct calls to __init__ do not clear previous
        # contents
        d = RankedDict([("a", 1), ("b", 2), ("c", 3), ("d", 44), ("e", 55)])
        d.__init__([("f", 6), ("e", 5)], d=4)
        self.assertEqual(
            list(d.items()),
            [("a", 1), ("b", 2), ("c", 3), ("f", 6), ("e", 5), ("d", 4)],
        )

    def test_update(self):
        with self.assertRaises(TypeError):
            # too many args
            RankedDict().update([("a", 1), ("b", 2)], None)
        pairs = [("a", 1), ("b", 2), ("c", 3), ("d", 4), ("e", 5)]
        od = RankedDict()
        od.update(dict(pairs))
        # dict input
        self.assertEqual(sorted(od.items()), pairs)
        od = RankedDict()
        od.update(**dict(pairs))
        # kwds input
        self.assertEqual(sorted(od.items()), pairs)
        od = RankedDict()
        od.update(pairs)
        # pairs input
        self.assertEqual(list(od.items()), pairs)
        od = RankedDict()
        od.update([("a", 1), ("b", 2), ("c", 9), ("d", 4)], c=3, e=5)
        # mixed input
        self.assertNotEqual(list(od.items()), pairs)

        # Issue 9137: Named argument called 'other' or 'self'
        # shouldn't be treated specially.
        od = RankedDict()
        od.update(self=23)
        self.assertEqual(list(od.items()), [("self", 23)])
        od = RankedDict()
        od.update(other={})
        self.assertEqual(list(od.items()), [("other", {})])
        od = RankedDict()
        od.update(red=5, blue=6, other=7, self=8)
        self.assertEqual(
            sorted(list(od.items())),
            [("blue", 6), ("other", 7), ("red", 5), ("self", 8)],
        )

    def test_abc(self):
        self.assertIsInstance(RankedDict(), MutableMapping)
        self.assertTrue(issubclass(RankedDict, MutableMapping))

    def test_clear(self):
        pairs = [("c", 1), ("b", 2), ("a", 3), ("d", 4), ("e", 5), ("f", 6)]
        shuffle(pairs)
        od = RankedDict(pairs)
        self.assertEqual(len(od), len(pairs))
        od.clear()
        self.assertEqual(len(od), 0)

    def test_delitem(self):
        pairs = [("c", 1), ("b", 2), ("a", 3), ("d", 4), ("e", 5), ("f", 6)]
        od = RankedDict(pairs)
        del od["a"]
        self.assertNotIn("a", od)
        with self.assertRaises(KeyError):
            del od["a"]
        self.assertEqual(list(od.items()), pairs[:2] + pairs[3:])

    def test_setitem(self):
        od = RankedDict([("d", 1), ("b", 2), ("c", 3), ("a", 4), ("e", 5)])
        od["c"] = 10  # existing element
        od["f"] = 20  # new element
        self.assertEqual(
            list(od.items()),
            [("d", 1), ("b", 2), ("a", 4), ("e", 5), ("c", 10), ("f", 20)],
        )

    def test_iterators(self):
        pairs = [("c", 1), ("b", 2), ("a", 3), ("d", 4), ("e", 5), ("f", 6)]
        shuffle(pairs)
        od = RankedDict(pairs)
        self.assertEqual(list(od), [t[0] for t in pairs])
        self.assertEqual(list(od.keys()), [t[0] for t in pairs])
        self.assertEqual(list(od.values()), [t[1] for t in pairs])
        self.assertEqual(list(od.items()), pairs)
        self.assertEqual(list(reversed(od)), [t[0] for t in reversed(pairs)])

    def test_popitem(self):
        pairs = [("c", 1), ("b", 2), ("a", 3), ("d", 4), ("e", 5), ("f", 6)]
        shuffle(pairs)
        od = RankedDict(pairs)
        while pairs:
            self.assertEqual(od.popitem(), pairs.pop())
        with self.assertRaises(KeyError):
            od.popitem()
        self.assertEqual(len(od), 0)

    def test_pop(self):
        pairs = [("c", 1), ("b", 2), ("a", 3), ("d", 4), ("e", 5), ("f", 6)]
        shuffle(pairs)
        od = RankedDict(pairs)
        shuffle(pairs)
        while pairs:
            k, v = pairs.pop()
            self.assertEqual(od.pop(k), v)
        with self.assertRaises(KeyError):
            od.pop("xyz")
        self.assertEqual(len(od), 0)
        self.assertEqual(od.pop(k, 12345), 12345)

        # make sure pop still works when __missing__ is defined
        class Missing(RankedDict):
            def __missing__(self, key):
                return 0

        m = Missing(a=1)
        self.assertEqual(m.pop("b", 5), 5)
        self.assertEqual(m.pop("a", 6), 1)
        self.assertEqual(m.pop("a", 6), 6)
        with self.assertRaises(KeyError):
            m.pop("a")

    def test_equality(self):
        pairs = [("c", 1), ("b", 2), ("a", 3), ("d", 4), ("e", 5), ("f", 6)]
        shuffle(pairs)
        od1 = RankedDict(pairs)
        od2 = RankedDict(pairs)
        # same order implies equality
        self.assertEqual(od1, od2)
        pairs = pairs[2:] + pairs[:2]
        od2 = RankedDict(pairs)
        # different order implies inequality
        self.assertNotEqual(od1, od2)
        # comparison to regular dict is not order sensitive
        self.assertEqual(od1, dict(od2))
        self.assertEqual(dict(od2), od1)
        # different length implied inequality
        self.assertNotEqual(od1, RankedDict(pairs[:-1]))

    def test_copying(self):
        # Check that ranked dicts are copyable, deepcopyable, picklable,
        # and have a repr/eval round-trip
        pairs = [("c", 1), ("b", 2), ("a", 3), ("d", 4), ("e", 5), ("f", 6)]
        od = RankedDict(pairs)
        update_test = RankedDict()
        update_test.update(od)
        for _i, dup in enumerate([
            od.copy(),
            copy.copy(od),
            copy.deepcopy(od),
            pickle.loads(pickle.dumps(od, 0)),
            pickle.loads(pickle.dumps(od, 1)),
            pickle.loads(pickle.dumps(od, 2)),
            pickle.loads(pickle.dumps(od, -1)),
            eval(repr(od)),
            update_test,
            RankedDict(od),
        ]):
            self.assertTrue(dup is not od)
            self.assertEqual(dup, od)
            self.assertEqual(list(dup.items()), list(od.items()))
            self.assertEqual(len(dup), len(od))
            self.assertEqual(type(dup), type(od))

    def test_yaml_linkage(self):
        # Verify that __reduce__ is setup in a way that supports PyYAML's
        # dump() feature.
        # In yaml, lists are native but tuples are not.
        pairs = [("c", 1), ("b", 2), ("a", 3), ("d", 4), ("e", 5), ("f", 6)]
        od = RankedDict(pairs)
        # yaml.dump(od) -->
        # '!!python/object/apply:__main__.RankedDict\n- - [a, 1]\n  - [b, 2]\n'
        self.assertTrue(all(type(pair) is list for pair in od.__reduce__()[1]))

    def test_repr(self):
        od = RankedDict([("c", 1), ("b", 2), ("a", 3)])
        self.assertEqual(repr(od), "RankedDict([('c', 1), ('b', 2), ('a', 3)])")
        self.assertEqual(eval(repr(od)), od)
        self.assertEqual(repr(RankedDict()), "RankedDict()")

    def test_repr_recursive(self):
        # See issue #9826
        od = RankedDict.fromkeys("abc")
        od["x"] = od
        self.assertEqual(
            repr(od),
            ("RankedDict([('a', None), ('b', None), " "('c', None), ('x', ...)])"),
        )

    def test_setdefault(self):
        pairs = [("c", 1), ("b", 2), ("a", 3), ("d", 4), ("e", 5), ("f", 6)]
        shuffle(pairs)
        od = RankedDict(pairs)
        pair_order = list(od.items())
        self.assertEqual(od.setdefault("a", 10), 3)
        # make sure order didn't change
        self.assertEqual(list(od.items()), pair_order)
        self.assertEqual(od.setdefault("x", 10), 10)
        # make sure 'x' is added to the end
        self.assertEqual(list(od.items())[-1], ("x", 10))

        # make sure setdefault still works when __missing__ is defined
        class Missing(RankedDict):
            def __missing__(self, key):
                return 0

        self.assertEqual(Missing().setdefault(5, 9), 9)

    def test_reinsert(self):
        # Given insert a, insert b, delete a, re-insert a,
        # verify that a is now later than b.
        od = RankedDict()
        od["a"] = 1
        od["b"] = 2
        del od["a"]
        od["a"] = 1
        self.assertEqual(list(od.items()), [("b", 2), ("a", 1)])

    def test_move_to_end(self):
        od = RankedDict.fromkeys("abcde")
        self.assertEqual(list(od), list("abcde"))
        od.move_to_end("c")
        self.assertEqual(list(od), list("abdec"))
        od.move_to_end("c", 0)
        self.assertEqual(list(od), list("cabde"))
        od.move_to_end("c", 0)
        self.assertEqual(list(od), list("cabde"))
        od.move_to_end("e")
        self.assertEqual(list(od), list("cabde"))
        with self.assertRaises(KeyError):
            od.move_to_end("x")

    @unittest.skipIf("PyPy" in sys.version, "sys.getsizeof not supported")
    def test_sizeof(self):
        # Wimpy test: Just verify the reported size is larger than a regular
        # dict
        d = dict(a=1)
        od = RankedDict(**d)
        self.assertGreater(sys.getsizeof(od), sys.getsizeof(d))


class TestPascalSet(unittest.TestCase):
    def test_consistency(self):
        from random import randint

        from xotl.tools.future.collections import PascalSet

        count = 5
        for test in range(count):
            size = randint(20, 60)
            ranges = (range(i, randint(i, i + 3)) for i in range(1, size))
            s1 = PascalSet(*ranges)
            ranges = (range(i, randint(i, i + 3)) for i in range(1, size))
            s2 = PascalSet(*ranges)
            ss1 = set(s1)
            ss2 = set(s2)
            self.assertEqual(s1, ss1)
            self.assertEqual(s1 - s2, ss1 - ss2)
            self.assertEqual(s2 - s1, ss2 - ss1)
            self.assertEqual(s1 & s2, ss1 & ss2)
            self.assertEqual(s2 & s1, ss2 & ss1)
            self.assertEqual(s1 | s2, ss1 | ss2)
            self.assertEqual(s2 | s1, ss2 | ss1)
            self.assertEqual(s1 ^ s2, ss1 ^ ss2)
            self.assertEqual(s2 ^ s1, ss2 ^ ss1)
            self.assertLess(s1 - s2, s1)
            self.assertLess(s1 - s2, ss1)
            self.assertLessEqual(s1 - s2, s1)
            self.assertLessEqual(s1 - s2, ss1)
            self.assertGreater(s1, s1 - s2)
            self.assertGreater(s1, ss1 - ss2)
            self.assertGreaterEqual(s1, s1 - s2)
            self.assertGreaterEqual(s1, ss1 - ss2)

    def test_syntax_sugar(self):
        from xotl.tools.future.collections import PascalSet

        s1 = PascalSet[1:4, 9, 15:18]
        s2 = PascalSet[3:18]
        self.assertEqual(str(s1), "{1..3, 9, 15..17}")
        self.assertEqual(str(s1 ^ s2), "{1, 2, 4..8, 10..14}")
        self.assertEqual(list(PascalSet[3:18]), list(range(3, 18)))

    def test_operators(self):
        from xotl.tools.future.collections import PascalSet

        g = lambda s: (i for i in s)
        s1 = PascalSet[1:4, 9, 15:18]
        r1 = range(1, 18)
        s2 = PascalSet(s1, 20)
        self.assertTrue(s1.issubset(s1))
        self.assertTrue(s1.issubset(set(s1)))
        self.assertTrue(s1.issubset(list(s1)))
        self.assertTrue(s1.issubset(g(s1)))
        self.assertTrue(s1.issubset(r1))
        self.assertTrue(s1.issubset(set(r1)))
        self.assertTrue(s1.issubset(list(r1)))
        self.assertTrue(s1.issubset(g(r1)))
        self.assertTrue(s2.issuperset(s2))
        self.assertTrue(s2.issuperset(s1))
        self.assertTrue(s2.issuperset(set(s1)))
        self.assertTrue(s2.issuperset(list(s1)))
        self.assertTrue(s2.issuperset(g(s1)))
        self.assertTrue(s1 <= set(s1))
        self.assertTrue(s1 < s2)
        self.assertTrue(s1 <= s2)
        self.assertTrue(s1 < set(s2))
        self.assertTrue(s1 <= set(s2))
        self.assertTrue(s1 < set(r1))
        self.assertTrue(s1 <= set(r1))
        self.assertTrue(s2 >= s2)
        self.assertTrue(s2 >= set(s2))
        self.assertTrue(s2 > s1)
        self.assertTrue(s2 > set(s1))
        self.assertTrue(s2 >= s1)
        self.assertTrue(s2 >= set(s1))

    def test_errors(self):
        """Test that stacked.pop has the same semantics has dict.pop."""
        from xotl.tools.future.collections import PascalSet

        s1 = PascalSet[1:4, 9, 15:18]
        s2 = PascalSet(s1, 20)
        self.assertLess(s1, s2)
        try:
            if s1 < list(s2):
                state = "less"
            else:
                state = "not-less"
        except TypeError:
            state = "TypeError"
        self.assertEqual(state, "TypeError")
        with self.assertRaises(TypeError):
            if s1 < set(s2):
                state = "ok"
            if s1 < list(s2):
                state = "safe-less"
            else:
                state = "safe-not-less"
        self.assertEqual(state, "ok")


class TestBitPascalSet(unittest.TestCase):
    def test_consistency(self):
        from random import randint

        from xotl.tools.future.collections import BitPascalSet

        count = 5
        for test in range(count):
            size = randint(20, 60)
            ranges = (range(i, randint(i, i + 3)) for i in range(1, size))
            s1 = BitPascalSet(*ranges)
            ranges = (range(i, randint(i, i + 3)) for i in range(1, size))
            s2 = BitPascalSet(*ranges)
            ss1 = set(s1)
            ss2 = set(s2)
            self.assertEqual(s1, ss1)
            self.assertEqual(s1 - s2, ss1 - ss2)
            self.assertEqual(s2 - s1, ss2 - ss1)
            self.assertEqual(s1 & s2, ss1 & ss2)
            self.assertEqual(s2 & s1, ss2 & ss1)
            self.assertEqual(s1 | s2, ss1 | ss2)
            self.assertEqual(s2 | s1, ss2 | ss1)
            self.assertEqual(s1 ^ s2, ss1 ^ ss2)
            self.assertEqual(s2 ^ s1, ss2 ^ ss1)
            self.assertLess(s1 - s2, s1)
            self.assertLess(s1 - s2, ss1)
            self.assertLessEqual(s1 - s2, s1)
            self.assertLessEqual(s1 - s2, ss1)
            self.assertGreater(s1, s1 - s2)
            self.assertGreater(s1, ss1 - ss2)
            self.assertGreaterEqual(s1, s1 - s2)
            self.assertGreaterEqual(s1, ss1 - ss2)

    def test_syntax_sugar(self):
        from xotl.tools.future.collections import BitPascalSet

        s1 = BitPascalSet[1:4, 9, 15:18]
        s2 = BitPascalSet[3:18]
        self.assertEqual(str(s1), "{1..3, 9, 15..17}")
        self.assertEqual(str(s1 ^ s2), "{1, 2, 4..8, 10..14}")
        self.assertEqual(list(BitPascalSet[3:18]), list(range(3, 18)))

    def test_operators(self):
        from xotl.tools.future.collections import BitPascalSet

        g = lambda s: (i for i in s)
        s1 = BitPascalSet[1:4, 9, 15:18]
        r1 = range(1, 18)
        s2 = BitPascalSet(s1, 20)
        self.assertTrue(s1.issubset(s1))
        self.assertTrue(s1.issubset(set(s1)))
        self.assertTrue(s1.issubset(list(s1)))
        self.assertTrue(s1.issubset(g(s1)))
        self.assertTrue(s1.issubset(r1))
        self.assertTrue(s1.issubset(set(r1)))
        self.assertTrue(s1.issubset(list(r1)))
        self.assertTrue(s1.issubset(g(r1)))
        self.assertTrue(s2.issuperset(s2))
        self.assertTrue(s2.issuperset(s1))
        self.assertTrue(s2.issuperset(set(s1)))
        self.assertTrue(s2.issuperset(list(s1)))
        self.assertTrue(s2.issuperset(g(s1)))
        self.assertTrue(s1 <= set(s1))
        self.assertTrue(s1 < s2)
        self.assertTrue(s1 <= s2)
        self.assertTrue(s1 < set(s2))
        self.assertTrue(s1 <= set(s2))
        self.assertTrue(s1 < set(r1))
        self.assertTrue(s1 <= set(r1))
        self.assertTrue(s2 >= s2)
        self.assertTrue(s2 >= set(s2))
        self.assertTrue(s2 > s1)
        self.assertTrue(s2 > set(s1))
        self.assertTrue(s2 >= s1)
        self.assertTrue(s2 >= set(s1))

    def test_errors(self):
        """Test that stacked.pop has the same semantics has dict.pop."""
        from xotl.tools.future.collections import BitPascalSet

        s1 = BitPascalSet[1:4, 9, 15:18]
        s2 = BitPascalSet(s1, 20)
        self.assertLess(s1, s2)
        try:
            if s1 < list(s2):
                state = "less"
            else:
                state = "not-less"
        except TypeError:
            state = "TypeError"
        self.assertEqual(state, "TypeError")
        with self.assertRaises(TypeError):
            if s1 < set(s2):
                state = "ok"
            if s1 < list(s2):
                state = "safe-less"
            else:
                state = "safe-not-less"
        self.assertEqual(state, "ok")


class TestCodeDict(unittest.TestCase):
    def test_formatter(self):
        from xotl.tools.future.collections import codedict

        cd = codedict(x=1, y=2, z=3.0)
        self.assertEqual(
            "{_[x + y]} is 3 --  {_[x + z]} is 4.0".format(_=cd),
            "3 is 3 --  4.0 is 4.0",
        )
        self.assertEqual(
            cd >> "{_[x + y]} is 3 --  {_[x + z]} is 4.0 -- {x} is 1",
            "3 is 3 --  4.0 is 4.0 -- 1 is 1",
        )
        self.assertEqual(
            "{_[x + y]} is 3 --  {_[x + z]} is 4.0 -- {x} is 1" << cd,
            "3 is 3 --  4.0 is 4.0 -- 1 is 1",
        )


def test_opendict():
    from enum import Enum

    from xotl.tools.future.collections import opendict

    class Foo:
        x = 1
        _y = 2

    foo = opendict.from_enum(Foo)
    assert dict(foo) == {"x": 1}

    class Bar(Enum):
        spam = "spam"

        def eat(self):
            return self.spam

    bar = opendict.from_enum(Bar)
    assert dict(bar) == {"spam": Bar.spam}


if __name__ == "__main__":
    # import sys;sys.argv = ['', 'Test.testName']
    unittest.main(verbosity=2)



# flake8: noqa

import numbers

from .meta import Dimension, Quantity

class Length(metaclass=Dimension):
    m: Quantity
    metre: Quantity
    meter: Quantity
    km: Quantity
    kilometer: Quantity
    kilometre: Quantity
    mm: Quantity
    centimetre: Quantity
    centimeter: Quantity
    cm: Quantity
    millimeter: Quantity
    millimetre: Quantity
    nm: Quantity
    nanometer: Quantity
    nanometre: Quantity

L = Length

class Time(metaclass=Dimension):
    s: Quantity
    second: Quantity
    ms: Quantity
    millisecond: Quantity
    ns: Quantity
    nanosecond: Quantity
    minute: Quantity
    hour: Quantity

T = Time

class Mass(metaclass=Dimension):
    kg: Quantity
    kilogram: Quantity
    gram: Quantity

M = Mass

class ElectricCurrent(metaclass=Dimension):
    A: Quantity
    ampere: Quantity
    milliampere: Quantity

I = ElectricCurrent

class Temperature(metaclass=Dimension):
    K: Quantity
    kelvin: Quantity
    @classmethod
    def from_celcius(cls, val: numbers.Real) -> Quantity: ...
    @classmethod
    def from_fahrenheit(cls, val: numbers.Real) -> Quantity: ...

O = Temperature

class Substance(metaclass=Dimension):
    mol: Quantity
    mole: Quantity

N = Substance

class Luminosity(metaclass=Dimension):
    candela: Quantity

J = Luminosity

class Area(metaclass=Dimension):
    metre_squared: Quantity
    meter_squared: Quantity

class Volumen(metaclass=Dimension):
    metre_cubic: Quantity
    meter_cubic: Quantity

class Frequency(metaclass=Dimension):
    unit_per_second: Quantity
    Hz: Quantity

class Force(metaclass=Dimension):
    metre_kilogram_per_second_squared: Quantity
    meter_kilogram_per_second_squared: Quantity
    N: Quantity
    Newton: Quantity

class Pressure(metaclass=Dimension):
    kilogram_per_metre_per_second_squared: Quantity
    Pascal: Quantity
    Pa: Quantity
    bar: Quantity

class Velocity(metaclass=Dimension):
    meter_per_second: Quantity
    metre_per_second: Quantity

class Acceleration(metaclass=Dimension):
    meter_per_second_squared: Quantity
    metre_per_second_squared: Quantity



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#

"""Decorator-making facilities.

This module provides a signature-keeping version of the
`xotl.tools.decorators.decorator`:func:, which is now deprecated in favor of
this module's version.

We scinded the decorator-making facilities from decorators per se to allow the
module `xotl.tools.deprecation`:mod: to be used by decorators and at the same
time, implement the decorator `~xotl.tools.deprecation.deprecated`:func: more
easily.

"""

from functools import partial, wraps
from types import FunctionType as function

__all__ = ("decorator",)


def decorator(caller):
    """Eases the creation of decorators with arguments.  Normally a decorator
    with arguments needs three nested functions like this::

        def decorator(*decorator_arguments):
            def real_decorator(target):
                def inner(*args, **kwargs):
                    return target(*args, **kwargs)
                return inner
            return real_decorator

    This decorator reduces the need of the first level by comprising both into
    a single function definition. However it does not removes the need for an
    ``inner`` function::

        >>> @decorator
        ... def plus(target, value):
        ...    from functools import wraps
        ...    @wraps(target)
        ...    def inner(*args):
        ...        return target(*args) + value
        ...    return inner

        >>> @plus(10)
        ... def ident(val):
        ...     return val

        >>> ident(1)
        11

    A decorator with default values for all its arguments (except, of course,
    the first one which is the decorated `target`) may be invoked
    without parenthesis::

        >>> @decorator
        ... def plus2(func, value=1, missing=2):
        ...    from functools import wraps
        ...    @wraps(func)
        ...    def inner(*args):
        ...        print(missing)
        ...        return func(*args) + value
        ...    return inner

        >>> @plus2
        ... def ident2(val):
        ...     return val

        >>> ident2(10)
        2
        11

    But (if you like) you may place the parenthesis::

        >>> @plus2()
        ... def ident3(val):
        ...     return val

        >>> ident3(10)
        2
        11

    However, this is not for free, you cannot pass a single positional argument
    which type is a function::

        >>> def p():
        ...    print('This is p!!!')

        >>> @plus2(p)   # doctest: +ELLIPSIS
        ... def dummy():
        ...    print('This is dummy')
        Traceback (most recent call last):
            ...
        TypeError: p() takes ...

    The workaround for this case is to use a keyword argument.

    """

    @wraps(caller)
    def outer_decorator(*args, **kwargs):
        try:
            from zope.interface import Interface
        except ImportError:
            Interface = None
            # from xotl.tools.symbols import Unset as Interface
        if (
            len(args) == 1
            and not kwargs
            and (
                isinstance(args[0], (function, type)) or issubclass(type(args[0]), type(Interface))
            )
        ):
            # This tries to solve the case of missing () on the decorator::
            #
            #    @decorator
            #    def somedec(func, *args, **kwargs)
            #        ...
            #
            #    @somedec
            #    def decorated(*args, **kwargs):
            #        pass
            #
            # Notice, however, that this is not general enough, since we try
            # to avoid inspecting the calling frame to see if the () are in
            # place.
            func = args[0]
            return caller(func)
        elif len(args) > 0 or len(kwargs) > 0:

            def _decorator(func):
                return partial(caller, **kwargs)(*((func,) + args))

            return _decorator
        else:
            return caller

    return outer_decorator



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
from hypothesis import given
from hypothesis import strategies as s


def test_slides():
    from xotl.tools.future.itertools import slides

    assert list(slides(range(1, 11))) == [
        (1, 2),
        (3, 4),
        (5, 6),
        (7, 8),
        (9, 10),
    ]


def test_slides_with_repeating_filling():
    from xotl.tools.future.itertools import NO_FILL, slides

    assert list(slides(range(1, 11), width=3, fill=None)) == [
        (1, 2, 3),
        (4, 5, 6),
        (7, 8, 9),
        (10, None, None),
    ]
    assert list(slides(range(1, 11), width=3, fill=NO_FILL)) == [
        (1, 2, 3),
        (4, 5, 6),
        (7, 8, 9),
        (10,),
    ]


def test_slides_with_cycling_filling():
    from xotl.tools.future.itertools import slides

    assert list(slides(range(1, 12), width=5, fill=(1, 2, 3))) == [
        (1, 2, 3, 4, 5),
        (6, 7, 8, 9, 10),
        (11, 1, 2, 3, 1),
    ]


def test_continuously_slides():
    from xotl.tools.future.itertools import continuously_slides

    aux = continuously_slides("maupasant", 3, "")
    trigrams = list("".join(x) for x in aux)
    assert "mau" in trigrams
    assert "aup" in trigrams
    assert "upa" in trigrams
    assert "pas" in trigrams
    assert "asa" in trigrams
    assert "san" in trigrams
    assert "ant" in trigrams
    assert len(trigrams) == 7


@s.composite
def keys(draw):
    return "k%d" % draw(s.integers(min_value=0, max_value=100))


@given(s.dictionaries(keys(), s.integers()), s.dictionaries(keys(), s.integers()))
def test_dict_update_new(d1, d2):
    from xotl.tools.future.itertools import dict_update_new

    d = dict(d1)
    dict_update_new(d1, d2)
    assert all(key in d1 for key in d2)
    assert all(d1[key] == d2[key] for key in d2 if key not in d)


@given(s.lists(s.integers(), max_size=30))
def test_delete_duplicates(ls):
    from collections import Counter

    from xotl.tools.future.itertools import delete_duplicates

    res = delete_duplicates(ls)
    assert type(ls) is type(res)  # noqa
    assert len(res) <= len(ls)
    assert all(Counter(res)[item] == 1 for item in ls)


@given(s.lists(s.integers(), max_size=30))
def test_delete_duplicates_with_key(ls):
    from xotl.tools.future.itertools import delete_duplicates

    res = delete_duplicates(ls, key=lambda x: x % 3)
    assert len(res) <= 3, "key yields 0, 1, or 2; thus res can contain at most 3 items"


def test_iter_delete_duplicates():
    from xotl.tools.future.itertools import iter_delete_duplicates

    assert list(iter_delete_duplicates("AAAaBBBA")) == ["A", "a", "B", "A"]
    assert list(iter_delete_duplicates("AAAaBBBA", key=lambda x: x.lower())) == [
        "A",
        "B",
        "A",
    ]

    assert list(iter_delete_duplicates("AAAaBBBA")) == ["A", "a", "B", "A"]
    assert list(iter_delete_duplicates("AAAaBBBA", key=lambda x: x.lower())) == ["A", "B", "A"]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# ---------------------------------------------------------------------
# Copyright (c) Merchise Autrement [~º/~] and Contributors
# All rights reserved.
#
# This is free software; you can do what the LICENCE file allows you to.
#
"""Enhanced data pretty printer."""

import io
from pprint import pprint

__all__ = ["ppformat"]


def ppformat(obj) -> str:
    """Just like `pprint`:func: but always returning a result.

    :returns: The pretty formated text.

    """
    stream = io.StringIO()
    pprint(obj, stream=stream)
    stream.seek(0)
    res = stream.read()
    return res



from xotl.tools.objects import classproperty, copy_class


class Foo:
    @classmethod
    def get_name(cls) -> str:
        return cls.__name__

    @classproperty
    def class_name(cls):
        return cls.get_name()


def check_str(s: str) -> str:
    return s


def check_foo(f: Foo) -> Foo:
    return f


if __name__ == "__main__":
    foo = Foo()
    check_str(foo.class_name)

    Foo2 = copy_class(Foo, new_name="Foo2")
    foo2 = Foo2()
    check_foo(foo2)
    check_str(foo2.class_name)



# flake8: noqa

from typing import (
    Any,
    Callable,
    ContextManager,
    Dict,
    Iterable,
    Iterator,
    List,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    Type,
    TypeVar,
    Union,
    overload,
)

from typing_extensions import Generic, Protocol
from xotl.tools.symbols import Unset

# def adapt_exception(value: Optional[Union[Tuple[Type[KeyError], str], str, int]], **kwargs) -> Optional[KeyError]: ...
Getter = Callable[[Any], Callable[[str, Any], Any]]
Setter = Callable[[Any], Callable[[str, Any], None]]

def dict_merge(*dicts: Dict, **others) -> Dict: ...
def extract_attrs(obj: Any, *names: str, default: Any = ...) -> Tuple[Any, ...]: ...
def fulldir(obj: Dict[Any, Any]) -> Set[str]: ...
def get_first_of(
    source: Any, *keys: str, default: Any = ..., pred: Callable[[Any], bool] = ...
) -> Optional[Any]: ...
def get_traverser(
    *paths: str, default: Any = ..., sep: str = ..., getter: Getter = ...
) -> Callable: ...
def traverse(
    obj: Any,
    path: str,
    default: Any = Unset,
    sep: str = ".",
    getter: Optional[Getter] = None,
) -> Any: ...
def import_object(
    name: Union[object, str],
    package: str = ...,
    sep: str = ...,
    default: Any = ...,
    **kwargs,
) -> Any: ...
def iterate_over(source: Any, *keys) -> Iterator[Any]: ...
def pop_first_of(source: Any, *keys: str, default: Any = ...) -> Any: ...
def popattr(obj: Any, name: str, default: Any = None) -> Any: ...
def save_attributes(
    obj: Any, *attrs: str, getter: Getter = ..., setter: Setter = ...
) -> ContextManager: ...
def smart_copy(*args, defaults=...) -> Dict[str, Any]: ...
def smart_getter(obj: Any, strict: bool = ...) -> Callable[[str, Any], Any]: ...
def smart_getter_and_deleter(obj: Any) -> Callable[[str, Any], Any]: ...
def smart_setter(obj: Any) -> Callable[[str, Any], None]: ...
def temp_attributes(
    obj: Any, attrs: Dict[Any, Any], getter: Getter = ..., setter: Setter = ...
) -> ContextManager: ...
def setdefaultattr(obj: Any, name: str, value: Any) -> Any: ...
def xdir(
    obj: Any,
    getter: Callable[[Any, str], Any],
    filter: Optional[Callable[[str, Any], bool]] = None,
    _depth: int = 0,
) -> Iterable[Tuple[str, Any]]: ...
def fdir(
    obj: Any,
    getter: Optional[Callable[[Any, str], Any]] = None,
    filter: Optional[Callable[[str, Any], bool]] = None,
) -> Iterable[str]: ...

class lazy:
    def __init__(self, value: Any, *args, **kawrgs) -> None: ...
    def __call__(self) -> Any: ...

_R_co = TypeVar("_R_co", covariant=True)

class classproperty(Generic[_R_co]):
    def __init__(self, method: Callable[[Any], _R_co]) -> None: ...
    def __get__(self, instance: Any, cls: Optional[type] = ...) -> _R_co: ...

memoized_property = property

T = TypeVar("T", bound=type)

@overload
def copy_class(
    cls: T,
    *,
    meta: Optional[Type[T]] = None,
    new_attrs: Optional[Mapping[str, Any]] = None,
    new_name: Optional[str] = None,
) -> T: ...
@overload
def copy_class(
    cls: T,
    *,
    meta: Optional[Type[T]] = None,
    new_attrs: Optional[Mapping[str, Any]] = None,
    new_name: Optional[str] = None,
    ignores: Sequence[str],
) -> Type: ...
def validate_attrs(
    source: Any,
    target: Any,
    force_equals: Sequence[str] = (),
    force_differents: Sequence[str] = (),
) -> bool: ...

class SafeDataItem:
    def __init__(self, *args, **kwargs) -> None: ...
    def __get__(self, obj, owner): ...
    @staticmethod
    def slot(slot_name: str, *args, **kwargs) -> str: ...

B = TypeVar("B")
B_co = TypeVar("B_co", covariant=True)

class _FinalSubclassEnum(Protocol[B]):
    def __getattr__(self, clsname: str) -> Type[B]: ...
    __members__: Mapping[str, Type[B]]

FinalSubclassEnumeration: _FinalSubclassEnum

class DelegatedAttribute:
    def __init__(self, target_name: str, delegated_attr: str, default: Any = Unset) -> None: ...
    def __get__(self, instance, owner) -> Any: ...

def delegator(attribute: str, attrs_map: Mapping[str, str], metaclass: type = type) -> type: ...
def iter_branch_subclasses(
    cls: Type[T],
    *,
    include_this: bool = False,
    without_duplicates: bool = False,
) -> Iterator[Type[T]]: ...
def get_branch_subclasses(
    cls: Type[T],
    *,
    include_this: bool = False,
    without_duplicates: bool = False,
) -> List[Type[T]]: ...
